import abc
import logging
import operator
from typing import Any, Callable, Literal, Optional, Sequence, Type, Union

import numpy as np
import scores
import xarray as xr
from scipy import ndimage

from extremeweatherbench import calc, derived, utils

logger = logging.getLogger(__name__)


class ComputeDocstringMetaclass(abc.ABCMeta):
    """A metaclass that maps the docstring from self._compute_metric() to
    self.compute_metric().

    The `BaseMetric` abstract base class requires users to override a function called
    `_compute_metric()`, while providing a standardized public interface to this method
    called `compute_metric()`. This metaclass automatically maps the docstring from
    `_compute_metric()` to `compute_metric()` so that the documentation a user provides
    for their implementation will automatically appear with the public interface without
    any additional effort.
    """

    def __new__(cls, name, bases, namespace):
        cls = super().__new__(cls, name, bases, namespace)
        # NOTE: the `compute_metric()` method will be defined in the ABC `BaseMetric`,
        # and we never expect the user re-implement it. So it won't be in the namespace
        # of the concrete metric classes - it will only be in the namespace of the ABC
        # `BaseMetric`, and will be available as an attribute of the concrete metric
        # classes.
        if "_compute_metric" in namespace and hasattr(cls, "compute_metric"):
            # Transfer the docstring from _compute_metric to compute_metric, if the
            # former exists.
            if cls._compute_metric.__doc__ is not None:
                # Create a new method for _this_ class, so we can avoid overwriting what
                # we set for the parent.
                _original_compute_metric = cls.compute_metric

                def _compute_metric_with_docstring(self, *args, **kwargs):
                    return _original_compute_metric(self, *args, **kwargs)

                _compute_metric_with_docstring.__doc__ = cls._compute_metric.__doc__
                cls.compute_metric = _compute_metric_with_docstring

        return cls


class BaseMetric(abc.ABC, metaclass=ComputeDocstringMetaclass):
    """Abstract base class defining the foundational interface for all metrics.

    Metrics are general operations applied between forecast and analysis xarray
    DataArrays. EWB metrics prioritize the use of any arbitrary sets of
    forecasts and analyses, so long as the spatiotemporal dimensions are the
    same.

    Public methods:
        compute_metric: Public interface to compute the metric
        maybe_expand_composite: Expand composite metrics into individual metrics
        is_composite: Check if this is a composite metric
        __repr__: String representation of the metric
        __eq__: Check equality with another metric

    Abstract methods:
        _compute_metric: Logic to compute the metric (must be implemented)
    """

    def __init__(
        self,
        name: str,
        preserve_dims: str = "lead_time",
        forecast_variable: Optional[str | derived.DerivedVariable] = None,
        target_variable: Optional[str | derived.DerivedVariable] = None,
    ):
        """Initialize the base metric.

        Args:
            name: The name of the metric.
            preserve_dims: The dimensions to preserve in the computation.
                Defaults to "lead_time".
            forecast_variable: The forecast variable to use in the
                computation.
            target_variable: The target variable to use in the computation.
        """
        # Store the original variables (str or DerivedVariable instances)
        # Do NOT convert to string to preserve output_variables info
        self.name = name
        self.preserve_dims = preserve_dims
        self.forecast_variable = forecast_variable
        self.target_variable = target_variable
        # Check if both variables are None - this is allowed
        if self.forecast_variable is None and self.target_variable is None:
            pass
        # If only one is None, raise an error
        elif self.forecast_variable is None or self.target_variable is None:
            raise ValueError(
                "Both forecast_variable and target_variable must be provided, "
                "or both must be None"
            )

    @abc.abstractmethod
    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> xr.DataArray:
        """Logic to compute, roll up, or otherwise transform the inputs for the base
        metric.

        All implementations must accept **kwargs to handle extra
        parameters gracefully, even if they don't use them.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.
            **kwargs: Additional parameters. Common ones include preserve_dims
                (dimension(s) to preserve, defaults to "lead_time").

        Returns:
            The computed metric result.
        """
        pass

    def compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Public interface to compute the metric.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.
            **kwargs: Additional keyword arguments to pass to the
                metric implementation.

        Returns:
            The computed metric result.
        """

        # If the forecast or target is sparse, densify it.
        # Ideally we would keep the sparse data structure, but Dask and sparse
        # do not play well together as of Nov 2025.
        forecast = utils.maybe_densify_dataarray(forecast)
        target = utils.maybe_densify_dataarray(target)
        return self._compute_metric(forecast, target, **kwargs)

    def maybe_expand_composite(self) -> Sequence["BaseMetric"]:
        """Expand composite metrics into individual metrics.

        Base implementation returns [self]. Override for composites.

        Returns:
            List containing just this metric.
        """
        return [self]

    def is_composite(self) -> bool:
        """Check if this is a composite metric.

        Base implementation returns False. Override for composites.

        Returns:
            False for base metrics.
        """
        return False

    def maybe_prepare_composite_kwargs(
        self,
        forecast_data: xr.DataArray,
        target_data: xr.DataArray,
        **base_kwargs: Any,
    ) -> dict:
        """Prepare kwargs for metric evaluation.

        Base implementation just returns kwargs as-is.
        Override for metrics that need special preparation.

        Args:
            forecast_data: The forecast DataArray.
            target_data: The target DataArray.

        Returns:
            Dictionary of kwargs (unchanged for base metrics).
        """
        return base_kwargs.copy()


class CompositeMetric(BaseMetric):
    """Base class for composite metrics that can contain multiple sub-metrics.

    Extends BaseMetric to provide functionality for composite metrics that
    aggregate multiple individual metrics for efficient evaluation.

    Public methods:
        maybe_expand_composite: Expand into individual metrics (overrides base)
        is_composite: Check if has sub-metrics (overrides base)

    Abstract methods:
        maybe_prepare_composite_kwargs: Prepare kwargs for composite evaluation
        _compute_metric: Compute the metric (must be implemented by subclasses)
    """

    def __init__(self, *args, **kwargs):
        """Initialize the composite metric.

        Args:
            *args: Positional arguments passed to BaseMetric.__init__
            **kwargs: Keyword arguments passed to BaseMetric.__init__
        """
        super().__init__(*args, **kwargs)
        self._metric_instances: list["BaseMetric"] = []

    def maybe_expand_composite(self) -> Sequence["BaseMetric"]:
        """Expand composite metrics into individual metrics.

        Returns:
            List containing just this metric.
        """
        if self._metric_instances:
            return self._metric_instances
        return [self]

    def is_composite(self) -> bool:
        """Check if this is a composite metric.

        Returns:
            True if composite (has sub-metrics), False otherwise.
        """
        return bool(self._metric_instances)

    @abc.abstractmethod
    def maybe_prepare_composite_kwargs(
        self,
        forecast_data: xr.DataArray,
        target_data: xr.DataArray,
        **base_kwargs,
    ) -> dict:
        """Prepare kwargs for composite metric evaluation.

        Returns:
            Dictionary of kwargs (unchanged for composite metrics).
        """

    def _compute_metric(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute metric (not supported for CompositeMetric base).

        CompositeMetric must be subclassed (like ThresholdMetric, LandfallMetric)
        or used as a composite with metrics list.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError(
            "CompositeMetric._compute_metric must be implemented "
            "by subclasses (ThresholdMetric, LandfallMetric) or use "
            "CompositeMetric as a composite with metrics=[...] list. Composites are "
            "automatically expanded in the evaluation pipeline."
        )


class ThresholdMetric(CompositeMetric):
    """Base class for threshold-based metrics with binary classification.

    Extends CompositeMetric to provide functionality for metrics that require
    forecast and target thresholds for binarization. Can be used as a base
    class for specific threshold metrics or as a composite metric.

    Public methods:
        transformed_contingency_manager: Create contingency manager
        maybe_prepare_composite_kwargs: Prepare kwargs (overrides parent)
        __call__: Make instances callable with configured thresholds

    Abstract methods:
        _compute_metric: Compute the metric (must be implemented by subclasses)

    Usage patterns:
        1. As a base class for specific metrics (CriticalSuccessIndex, etc.)
        2. As a composite metric to compute multiple threshold metrics
           efficiently by reusing the transformed contingency manager

    Example:
        composite = ThresholdMetric(
            metrics=[CriticalSuccessIndex, FalseAlarmRatio, Accuracy],
            forecast_threshold=0.7,
            target_threshold=0.5
        )
    """

    def __init__(
        self,
        name: str = "threshold_metrics",
        preserve_dims: str = "lead_time",
        forecast_variable: Optional[str | derived.DerivedVariable] = None,
        target_variable: Optional[str | derived.DerivedVariable] = None,
        forecast_threshold: float = 0.5,
        target_threshold: float = 0.5,
        metrics: Optional[list[Type["ThresholdMetric"]]] = None,
        **kwargs,
    ):
        """Initialize the threshold metric.

        Args:
            name: The name of the metric. Defaults to "threshold_metrics".
            preserve_dims: The dimensions to preserve in the computation.
                Defaults to "lead_time".
            forecast_variable: The forecast variable to use in the
                computation.
            target_variable: The target variable to use in the computation.
            forecast_threshold: The threshold for binarizing the forecast.
                Defaults to 0.5.
            target_threshold: The threshold for binarizing the target.
                Defaults to 0.5.
            metrics: A list of metrics to use as a composite. Defaults to
                None.
            **kwargs: Additional keyword arguments passed to parent.
        """
        super().__init__(
            name,
            preserve_dims=preserve_dims,
            forecast_variable=forecast_variable,
            target_variable=target_variable,
            **kwargs,
        )
        self.forecast_threshold = forecast_threshold
        self.target_threshold = target_threshold
        self.preserve_dims = preserve_dims
        self.metrics = metrics or []

        # If metrics provided, instantiate them
        if self.metrics is not None:
            self._metric_instances = [
                (
                    metric_cls(
                        forecast_threshold=self.forecast_threshold,
                        target_threshold=self.target_threshold,
                        preserve_dims=self.preserve_dims,
                    )
                    if isinstance(metric_cls, type)
                    else metric_cls
                )
                for metric_cls in self.metrics
            ]
        else:
            self._metric_instances = []

    def __call__(self, forecast: xr.DataArray, target: xr.DataArray, **kwargs) -> Any:
        """Make instances callable using their configured thresholds."""
        # Use instance attributes as defaults, but allow override from kwargs
        kwargs.setdefault("forecast_threshold", self.forecast_threshold)
        kwargs.setdefault("target_threshold", self.target_threshold)
        kwargs.setdefault("preserve_dims", self.preserve_dims)

        # Call the instance method with the configured parameters
        return self.compute_metric(forecast, target, **kwargs)

    def transformed_contingency_manager(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        forecast_threshold: float,
        target_threshold: float,
        preserve_dims: str,
        op_func: Union[
            Callable, Literal[">", ">=", "<", "<=", "==", "!="]
        ] = operator.ge,
    ) -> scores.categorical.BasicContingencyManager:
        """Create and transform a contingency manager.

        This method is used to create and transform a contingency manager from the
        scores module. The op_func is used to binarize the forecast and target data with
        either a string representation of the operator, e.g. ">=", or a callable
        function from the operator module, e.g. operator.ge.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.
            forecast_threshold: Threshold for binarizing forecast.
            target_threshold: Threshold for binarizing target.
            preserve_dims: Dimension(s) to preserve during transform.
            op_func: Function or string representation of the operator to apply to the
                forecast and target. Defaults to operator.ge (greater than or equal to).

        Returns:
            Transformed contingency manager.
        """
        # Apply thresholds to binarize the data
        op_func = utils.maybe_get_operator(op_func)
        binary_forecast = utils.maybe_densify_dataarray(
            op_func(forecast, forecast_threshold)
        ).astype(float)
        binary_target = utils.maybe_densify_dataarray(
            op_func(target, target_threshold)
        ).astype(float)

        # Create and transform contingency manager
        binary_contingency_manager = scores.categorical.BinaryContingencyManager(
            binary_forecast, binary_target
        )
        transformed = binary_contingency_manager.transform(preserve_dims=preserve_dims)

        return transformed

    def maybe_prepare_composite_kwargs(
        self,
        forecast_data: xr.DataArray,
        target_data: xr.DataArray,
        **base_kwargs: Any,
    ) -> dict:
        """Prepare kwargs for composite metric evaluation.

        Computes the transformed contingency manager once and adds
        it to kwargs for efficient composite evaluation.

        Args:
            forecast_data: The forecast DataArray.
            target_data: The target DataArray.

        Returns:
            Dictionary of kwargs including transformed_manager.
        """
        kwargs = base_kwargs.copy()

        if self.is_composite() and len(self._metric_instances) > 1:
            kwargs["transformed_manager"] = self.transformed_contingency_manager(
                forecast=forecast_data,
                target=target_data,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
            kwargs["forecast_threshold"] = self.forecast_threshold
            kwargs["target_threshold"] = self.target_threshold
            kwargs["preserve_dims"] = self.preserve_dims

        return kwargs

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute metric (not supported for ThresholdMetric base).

        ThresholdMetric must be subclassed (like CriticalSuccessIndex, FalseAlarmRatio)
        or used as a composite with metrics list.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError(
            "ThresholdMetric._compute_metric must be implemented "
            "by subclasses (CriticalSuccessIndex, FalseAlarmRatio, etc.) or use "
            "ThresholdMetric as a composite with metrics=[...] list. Composites are "
            "automatically expanded in the evaluation pipeline."
        )


class CriticalSuccessIndex(ThresholdMetric):
    """Compute Critical Success Index (CSI) from binary classifications.

    Extends ThresholdMetric to compute CSI between forecast and target using
    the preserve_dims dimensions. CSI measures the fraction of correctly
    predicted events.
    """

    def __init__(self, name: str = "CriticalSuccessIndex", *args, **kwargs):
        """Initialize the Critical Success Index metric.

        Args:
            name: The name of the metric. Defaults to
                "CriticalSuccessIndex".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        return transformed.critical_success_index()


class FalseAlarmRatio(ThresholdMetric):
    """Compute False Alarm Ratio (FAR) from binary classifications.

    Extends ThresholdMetric to compute FAR between forecast and target using
    the preserve_dims dimensions. FAR measures the fraction of predicted
    events that did not occur. Note: FAR is not the same as False Alarm Rate.
    """

    def __init__(self, name: str = "FalseAlarmRatio", *args, **kwargs):
        """Initialize the False Alarm Ratio metric.

        Args:
            name: The name of the metric. Defaults to "FalseAlarmRatio".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        return transformed.false_alarm_ratio()


class TruePositives(ThresholdMetric):
    """Compute True Positive ratio from binary classifications.

    Extends ThresholdMetric to compute the ratio of true positives (correctly
    predicted events) to the total number of observations. Corresponds to the
    top right cell in the contingency table.
    """

    def __init__(self, name: str = "TruePositives", *args, **kwargs):
        """Initialize the True Positives metric.

        Args:
            name: The name of the metric. Defaults to "TruePositives".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        counts = transformed.get_counts()
        return counts["tp_count"] / counts["total_count"]


class FalsePositives(ThresholdMetric):
    """Compute False Positive ratio from binary classifications.

    Extends ThresholdMetric to compute the ratio of false positives
    (incorrectly predicted events) to the total number of observations.
    """

    def __init__(self, name: str = "FalsePositives", *args, **kwargs):
        """Initialize the False Positives metric.

        Args:
            name: The name of the metric. Defaults to "FalsePositives".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        counts = transformed.get_counts()
        return counts["fp_count"] / counts["total_count"]


class TrueNegatives(ThresholdMetric):
    """Compute True Negative ratio from binary classifications.

    Extends ThresholdMetric to compute the ratio of true negatives (correctly
    predicted non-events) to the total number of observations.
    """

    def __init__(self, name: str = "TrueNegatives", *args, **kwargs):
        """Initialize the True Negatives metric.

        Args:
            name: The name of the metric. Defaults to "TrueNegatives".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        counts = transformed.get_counts()
        return counts["tn_count"] / counts["total_count"]


class FalseNegatives(ThresholdMetric):
    """Compute False Negative ratio from binary classifications.

    Extends ThresholdMetric to compute the ratio of false negatives (missed
    events) to the total number of observations. Corresponds to the top left
    cell in the contingency table.
    """

    def __init__(self, name: str = "FalseNegatives", *args, **kwargs):
        """Initialize the False Negatives metric.

        Args:
            name: The name of the metric. Defaults to "FalseNegatives".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        counts = transformed.get_counts()
        return counts["fn_count"] / counts["total_count"]


class Accuracy(ThresholdMetric):
    """Compute classification accuracy from binary classifications.

    Extends ThresholdMetric to compute the ratio of correct predictions (true
    positives + true negatives) to the total number of observations. Measures
    overall correctness of the forecast.
    """

    def __init__(self, name: str = "Accuracy", *args, **kwargs):
        """Initialize the Accuracy metric.

        Args:
            name: The name of the metric. Defaults to "Accuracy".
            *args: Additional positional arguments passed to ThresholdMetric.
            **kwargs: Additional keyword arguments passed to ThresholdMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        # Use pre-computed manager if provided, else compute
        transformed = kwargs.get("transformed_manager")
        if transformed is None:
            transformed = self.transformed_contingency_manager(
                forecast=forecast,
                target=target,
                forecast_threshold=self.forecast_threshold,
                target_threshold=self.target_threshold,
                preserve_dims=self.preserve_dims,
            )
        return transformed.accuracy()


class MeanSquaredError(BaseMetric):
    """Compute Mean Squared Error between forecast and target.

    Extends BaseMetric to calculate MSE with optional interval-based
    weighting and custom weights for spatial/temporal averaging.
    """

    def __init__(
        self,
        name: str = "MeanSquaredError",
        interval_where_one: Optional[
            tuple[int | float | xr.DataArray, int | float | xr.DataArray]
        ] = None,
        interval_where_positive: Optional[
            tuple[int | float | xr.DataArray, int | float | xr.DataArray]
        ] = None,
        weights: Optional[xr.DataArray] = None,
        *args,
        **kwargs,
    ):
        """Initialize the Mean Squared Error metric.

        Args:
            name: The name of the metric. Defaults to "MeanSquaredError".
            interval_where_one: Endpoints of the interval where threshold
                weights are 1. Must be increasing. Infinite endpoints
                permissible.
            interval_where_positive: Endpoints of the interval where threshold
                weights are positive. Must be increasing.
            weights: Array of weights to apply to the score (e.g., latitude
                weighting). If None, no weights are applied.
            *args: Additional positional arguments passed to BaseMetric.
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        super().__init__(name, *args, **kwargs)
        self.interval_where_one = interval_where_one
        self.interval_where_positive = interval_where_positive
        self.weights = weights

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        if self.interval_where_one is not None:
            return scores.continuous.tw_squared_error(
                forecast,
                target,
                interval_where_one=self.interval_where_one,
                interval_where_positive=self.interval_where_positive,
                weights=self.weights,
                preserve_dims=self.preserve_dims,
            )
        return scores.continuous.mse(forecast, target, preserve_dims=self.preserve_dims)


class MeanAbsoluteError(BaseMetric):
    """Compute Mean Absolute Error between forecast and target.

    Extends BaseMetric to calculate MAE with optional interval-based
    weighting and custom weights for spatial/temporal averaging.
    """

    def __init__(
        self,
        name: str = "MeanAbsoluteError",
        interval_where_one: Optional[
            tuple[int | float | xr.DataArray, int | float | xr.DataArray]
        ] = None,
        interval_where_positive: Optional[
            tuple[int | float | xr.DataArray, int | float | xr.DataArray]
        ] = None,
        weights: Optional[xr.DataArray] = None,
        *args,
        **kwargs,
    ):
        """Initialize the Mean Absolute Error metric.

        Args:
            name: The name of the metric. Defaults to "MeanAbsoluteError".
            interval_where_one: Endpoints of the interval where threshold
                weights are 1. Must be increasing. Infinite endpoints
                permissible.
            interval_where_positive: Endpoints of the interval where threshold
                weights are positive. Must be increasing.
            weights: Array of weights to apply to the score (e.g., latitude
                weighting). If None, no weights are applied.
            *args: Additional positional arguments passed to BaseMetric.
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        self.interval_where_one = interval_where_one
        self.interval_where_positive = interval_where_positive
        self.weights = weights
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute the Mean Absolute Error.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            The computed Mean Absolute Error result.
        """
        if self.interval_where_one is not None:
            return scores.continuous.tw_absolute_error(
                forecast,
                target,
                interval_where_one=self.interval_where_one,
                interval_where_positive=self.interval_where_positive,
                weights=self.weights,
                preserve_dims=self.preserve_dims,
            )
        return scores.continuous.mae(forecast, target, preserve_dims=self.preserve_dims)


class MeanError(BaseMetric):
    """Compute Mean Error (bias) between forecast and target.

    Extends BaseMetric to calculate mean error (bias) using the preserve_dims
    dimensions. Positive values indicate forecast exceeds target.
    """

    def __init__(self, name: str = "MeanError", *args, **kwargs):
        """Initialize the Mean Error metric.

        Args:
            name: The name of the metric. Defaults to "MeanError".
            *args: Additional positional arguments passed to BaseMetric.
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute the Mean Error.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            The computed Mean Error result.
        """
        return scores.continuous.mean_error(
            forecast, target, preserve_dims=self.preserve_dims
        )


class RootMeanSquaredError(BaseMetric):
    """Compute Root Mean Squared Error between forecast and target.

    Extends BaseMetric to calculate RMSE using the preserve_dims dimensions.
    RMSE is the square root of the mean squared error.
    """

    def __init__(self, name: str = "RootMeanSquaredError", *args, **kwargs):
        """Initialize the Root Mean Squared Error metric.

        Args:
            name: The name of the metric. Defaults to "RootMeanSquaredError".
            *args: Additional positional arguments passed to BaseMetric.
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute the Root Mean Square Error.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            The computed Root Mean Square Error result.
        """
        return scores.continuous.rmse(
            forecast, target, preserve_dims=self.preserve_dims
        )


class EarlySignal(BaseMetric):
    """Detect first occurrence of signal exceeding threshold criteria.

    Extends BaseMetric to find the earliest time when a signal is detected
    based on threshold criteria, returning init_time, lead_time, and
    valid_time information. Flexible for different signal detection criteria.
    """

    def __init__(
        self,
        name: str = "EarlySignal",
        comparison_operator: Union[
            Callable, Literal[">", ">=", "<", "<=", "==", "!="]
        ] = ">=",
        threshold: float = 0.5,
        spatial_aggregation: Literal["any", "all", "half"] = "any",
        **kwargs,
    ):
        """Initialize the Early Signal detection metric.

        Args:
            name: The name of the metric. Defaults to "EarlySignal".
            comparison_operator: The comparison operator for signal detection.
            threshold: The threshold value for signal detection.
            spatial_aggregation: Spatial aggregation method. Options: "any"
                (any gridpoint meets criteria), "all" (all gridpoints meet
                criteria), or "half" (at least half meet criteria).
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        # Extract threshold params before passing to super
        self.comparison_operator = utils.maybe_get_operator(comparison_operator)
        self.threshold = threshold
        self.spatial_aggregation = spatial_aggregation
        super().__init__(name, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> xr.DataArray:
        """Compute early signal detection.

        Args:
            forecast: The forecast dataarray with init_time, lead_time, valid_time.
            target: The target dataarray (used for reference/validation).

        Returns:
            Boolean DataArray with dims [init_time, lead_time] indicating
            whether criteria are met for each init_time and lead_time pair.
        """
        if self.threshold is None:
            # Return False for all when no detection criteria specified
            dims = ["init_time", "lead_time"]
            coords = {
                "init_time": forecast.valid_time - forecast.lead_time,
                "lead_time": forecast.lead_time,
            }
            if "valid_time" in forecast.dims:
                dims.append("valid_time")
                coords["valid_time"] = forecast.valid_time
            return xr.DataArray(
                False,
                dims=dims,
                coords=coords,
                name=self.name,
            )
        # Create detection mask
        detection_mask = self.comparison_operator(forecast, self.threshold)

        # Apply spatial aggregation
        spatial_dims = [
            dim
            for dim in detection_mask.dims
            if dim not in ["init_time", "lead_time", "valid_time"]
        ]

        if spatial_dims:
            if self.spatial_aggregation == "any":
                detection_mask = detection_mask.any(spatial_dims)
            elif self.spatial_aggregation == "all":
                detection_mask = detection_mask.all(spatial_dims)
            elif self.spatial_aggregation == "half":
                detection_mask = operator.ge(detection_mask.mean(spatial_dims), 0.5)
            else:
                raise ValueError(
                    f"Spatial aggregation '{self.spatial_aggregation}' not supported"
                )

        detection_mask.name = self.name
        return detection_mask


class MaximumMeanAbsoluteError(MeanAbsoluteError):
    """Compute MAE between forecast and target maximum values.

    Extends MeanAbsoluteError to filter forecast to a time window around the
    target's maximum using tolerance_range_hours. Useful for evaluating peak
    value timing and magnitude.
    """

    def __init__(
        self,
        tolerance_range_hours: int = 24,
        reduce_spatial_dims: list[str] = ["latitude", "longitude"],
        name: str = "MaximumMeanAbsoluteError",
        *args,
        **kwargs,
    ):
        """Initialize the Maximum Mean Absolute Error metric.

        Args:
            tolerance_range_hours: Time window (hours) around target's
                maximum to search for forecast maximum. Defaults to 24.
            reduce_spatial_dims: Spatial dimensions to reduce. Defaults to
                ["latitude", "longitude"].
            name: The name of the metric. Defaults to
                "MaximumMeanAbsoluteError".
            *args: Additional positional arguments passed to
                MeanAbsoluteError.
            **kwargs: Additional keyword arguments passed to
                MeanAbsoluteError.
        """
        self.tolerance_range_hours = tolerance_range_hours
        self.reduce_spatial_dims = reduce_spatial_dims
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> xr.DataArray:
        """Compute MaximumMeanAbsoluteError.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            MeanAbsoluteError of the maximum values.
        """
        # Enforced spatial reduction for MaximumMeanAbsoluteError
        reduce_spatial_dims = ["latitude", "longitude"]
        target_spatial_mean = utils.reduce_dataarray(
            target, method="mean", reduce_dims=reduce_spatial_dims, skipna=True
        )
        maximum_timestep = target_spatial_mean.idxmax("valid_time")
        maximum_value = target_spatial_mean.sel(valid_time=maximum_timestep)

        # Handle the case where there are >1 resulting target values
        maximum_timestep = utils.maybe_get_closest_timestamp_to_center_of_valid_times(
            maximum_timestep, target.valid_time
        ).compute()
        forecast_spatial_mean = utils.reduce_dataarray(
            forecast, method="mean", reduce_dims=reduce_spatial_dims, skipna=True
        )
        filtered_max_forecast = forecast_spatial_mean.where(
            (
                forecast_spatial_mean.valid_time
                >= maximum_timestep.data
                - np.timedelta64(self.tolerance_range_hours // 2, "h")
            )
            & (
                forecast_spatial_mean.valid_time
                <= maximum_timestep.data
                + np.timedelta64(self.tolerance_range_hours // 2, "h")
            ),
            drop=True,
        ).max("valid_time")
        return super()._compute_metric(
            forecast=filtered_max_forecast,
            target=maximum_value,
            preserve_dims=self.preserve_dims,
        )


class MinimumMeanAbsoluteError(MeanAbsoluteError):
    """Compute MAE between forecast and target minimum values.

    Extends MeanAbsoluteError to filter forecast to a time window around the
    target's minimum using tolerance_range_hours. Useful for evaluating
    minimum value timing and magnitude.
    """

    def __init__(
        self,
        tolerance_range_hours: int = 24,
        reduce_spatial_dims: list[str] = ["latitude", "longitude"],
        name: str = "MinimumMeanAbsoluteError",
        *args,
        **kwargs,
    ):
        """Initialize the Minimum Mean Absolute Error metric.

        Args:
            tolerance_range_hours: Time window (hours) around target's
                minimum to search for forecast minimum. Defaults to 24.
            reduce_spatial_dims: Spatial dimensions to reduce. Defaults to
                ["latitude", "longitude"].
            name: The name of the metric. Defaults to
                "MinimumMeanAbsoluteError".
            *args: Additional positional arguments passed to
                MeanAbsoluteError.
            **kwargs: Additional keyword arguments passed to
                MeanAbsoluteError.
        """
        self.tolerance_range_hours = tolerance_range_hours
        self.reduce_spatial_dims = reduce_spatial_dims
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute MinimumMeanAbsoluteError.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            MeanAbsoluteError of the minimum values.
        """
        target_spatial_mean = utils.reduce_dataarray(
            target, method="mean", reduce_dims=self.reduce_spatial_dims, skipna=True
        )
        minimum_timestep = target_spatial_mean.idxmin("valid_time")
        minimum_value = target_spatial_mean.sel(valid_time=minimum_timestep)
        forecast_spatial_mean = utils.reduce_dataarray(
            forecast, method="mean", reduce_dims=self.reduce_spatial_dims, skipna=True
        )
        # Handle the case where there are >1 resulting target values
        minimum_timestep = utils.maybe_get_closest_timestamp_to_center_of_valid_times(
            minimum_timestep, target.valid_time
        )
        filtered_min_forecast = forecast_spatial_mean.where(
            (
                forecast_spatial_mean.valid_time
                >= minimum_timestep.data
                - np.timedelta64(self.tolerance_range_hours // 2, "h")
            )
            & (
                forecast_spatial_mean.valid_time
                <= minimum_timestep.data
                + np.timedelta64(self.tolerance_range_hours // 2, "h")
            ),
            drop=True,
        ).min("valid_time")
        return super()._compute_metric(
            forecast=filtered_min_forecast,
            target=minimum_value,
            preserve_dims=self.preserve_dims,
        )


class MaximumLowestMeanAbsoluteError(MeanAbsoluteError):
    """Compute MAE of maximum aggregated minimum values for heatwaves.

    Extends MeanAbsoluteError for heatwave evaluation by aggregating daily
    minimum values and computing MAE between the warmest nighttime (daily
    minimum) temperature in target and forecast.
    """

    def __init__(
        self,
        tolerance_range_hours: int = 24,
        name: str = "MaximumLowestMeanAbsoluteError",
        *args,
        **kwargs,
    ):
        """Initialize the Maximum Lowest Mean Absolute Error metric.

        Args:
            tolerance_range_hours: Time window (hours) around target's
                max-min value to search for forecast max-min. Defaults to 24.
            name: The name of the metric. Defaults to
                "MaximumLowestMeanAbsoluteError".
            *args: Additional positional arguments passed to
                MeanAbsoluteError.
            **kwargs: Additional keyword arguments passed to
                MeanAbsoluteError.
        """
        self.tolerance_range_hours = tolerance_range_hours
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute MaximumLowestMeanAbsoluteError.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            MeanAbsoluteError of the highest aggregated minimum value.
        """
        reduce_dims = [
            dim
            for dim in forecast.dims
            if dim not in ["valid_time", "lead_time", "time"]
        ]
        forecast = utils.reduce_dataarray(
            forecast, method="mean", reduce_dims=reduce_dims, skipna=True
        )
        target = utils.reduce_dataarray(
            target, method="mean", reduce_dims=reduce_dims, skipna=True
        )

        time_resolution_hours = utils.determine_temporal_resolution(target)
        max_min_target_value = (
            target.groupby("valid_time.dayofyear")
            .map(
                utils.min_if_all_timesteps_present,
                time_resolution_hours=time_resolution_hours,
            )
            .max()
        )
        max_min_target_datetime = target.where(
            target == max_min_target_value, drop=True
        ).valid_time

        # Handle the case where there are >1 resulting target values
        max_min_target_datetime = (
            utils.maybe_get_closest_timestamp_to_center_of_valid_times(
                max_min_target_datetime, target.valid_time
            )
        )
        subset_forecast = (
            forecast.where(
                (
                    forecast.valid_time
                    >= (
                        max_min_target_datetime.data
                        - np.timedelta64(self.tolerance_range_hours // 2, "h")
                    )
                )
                & (
                    forecast.valid_time
                    <= (
                        max_min_target_datetime.data
                        + np.timedelta64(self.tolerance_range_hours // 2, "h")
                    )
                ),
                drop=True,
            )
            .groupby("valid_time.dayofyear")
            .map(
                utils.min_if_all_timesteps_present_forecast,
                time_resolution_hours=utils.determine_temporal_resolution(forecast),
            )
            .min("dayofyear")
        )

        return super()._compute_metric(
            forecast=subset_forecast,
            target=max_min_target_value,
            preserve_dims=self.preserve_dims,
        )


class DurationMeanError(MeanError):
    """Compute mean error of event duration between forecast and target.

    Extends MeanError to compute the mean error between forecast and target
    event durations based on threshold criteria and spatial aggregation.
    """

    def __init__(
        self,
        threshold_criteria: xr.DataArray | float,
        reduce_spatial_dims: list[str] = ["latitude", "longitude"],
        op_func: Union[Callable, Literal[">", ">=", "<", "<=", "==", "!="]] = ">=",
        name: str = "DurationMeanError",
        preserve_dims: str = "init_time",
        product_time_resolution_hours: bool = False,
    ):
        """Initialize the Duration Mean Error metric.

        Args:
            threshold_criteria: Criteria for event detection. Either a
                DataArray of climatology with dimensions (dayofyear, hour,
                latitude, longitude) or a float fixed threshold.
            reduce_spatial_dims: Spatial dimensions to reduce prior to
                applying threshold criteria. Defaults to ["latitude",
                "longitude"].
            op_func: Comparison operator or string (e.g., operator.ge for
                >=).
            name: Name of the metric. Defaults to "DurationMeanError".
            preserve_dims: Dimensions to preserve during aggregation.
                Defaults to "init_time".
            product_time_resolution_hours: Whether to multiply duration by
                time resolution of forecast (in hours). Defaults to False.
        """
        super().__init__(name=name, preserve_dims=preserve_dims)
        self.reduce_spatial_dims = reduce_spatial_dims
        self.threshold_criteria = threshold_criteria
        self.op_func = utils.maybe_get_operator(op_func)
        self.product_time_resolution_hours = product_time_resolution_hours

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs,
    ) -> Any:
        """Compute spatially averaged duration mean error.

        Args:
            forecast: the forecast DataArray.
            target: the target DataArray.

        Returns:
            The mean error between forecast and target event durations.
        """
        # Handle criteria - either climatology (xr.DataArray) or float threshold
        # Use local variable to avoid mutating self.threshold_criteria
        threshold_criteria = self.threshold_criteria

        # Need to get climatology into the correct format and interpolation for
        # comparison
        if isinstance(threshold_criteria, xr.DataArray):
            # Climatology case, convert from dayofyear/hour to valid_time.
            # Note that unintended behavior may occur if the case spans multiple years.
            threshold_criteria = utils.convert_day_yearofday_to_time(
                threshold_criteria, forecast.valid_time.dt.year.values[0]
            )

            # Interpolate climatology to target coordinates
            threshold_criteria = utils.interp_climatology_to_target(
                target, threshold_criteria
            )
        # Reduce spatial dimensions if specified (default is to reduce)
        if len(self.reduce_spatial_dims) > 0:
            target = utils.reduce_dataarray(
                target, method="mean", reduce_dims=self.reduce_spatial_dims, skipna=True
            )
            forecast = utils.reduce_dataarray(
                forecast,
                method="mean",
                reduce_dims=self.reduce_spatial_dims,
                skipna=True,
            )

            if isinstance(threshold_criteria, xr.DataArray):
                threshold_criteria = utils.reduce_dataarray(
                    threshold_criteria,
                    method="mean",
                    reduce_dims=self.reduce_spatial_dims,
                    skipna=True,
                )
        forecast_mask = self.op_func(forecast, threshold_criteria)
        target_mask = self.op_func(target, threshold_criteria)

        # Track NaN locations in forecast data
        forecast_valid_mask = ~forecast.isnull()

        # Apply valid data mask (exclude NaN positions in forecast)
        forecast_mask_final = forecast_mask.where(forecast_valid_mask)
        try:
            target_mask_final = target_mask.where(forecast_valid_mask)
        # If sparse, will need to expand_dims first as transpose is not supported
        except AttributeError:
            logger.info(
                "Target mask is sparse, expanding dimensions to handle unsupported "
                "transpose operation."
            )
            target_mask_final = target_mask.expand_dims(
                dim={"lead_time": target.lead_time.size}
            ).where(forecast_valid_mask)

        # Sum to get durations (NaN values are excluded by default)
        forecast_duration = forecast_mask_final.groupby(self.preserve_dims).sum(
            skipna=True
        )
        target_duration = target_mask_final.groupby(self.preserve_dims).sum(skipna=True)

        if self.product_time_resolution_hours:
            time_resolution_hours = utils.determine_temporal_resolution(forecast)
            forecast_duration = forecast_duration * time_resolution_hours
            target_duration = target_duration * time_resolution_hours

        return super()._compute_metric(
            forecast=forecast_duration,
            target=target_duration,
            preserve_dims=self.preserve_dims,
        )


class LandfallMetric(CompositeMetric):
    """Base class for tropical cyclone landfall metrics.

    Extends CompositeMetric to compute landfalls using calc.find_landfalls,
    which utilizes land geometry and line segments based on track data to
    determine intersections.

    Can be used as a base class for custom landfall metrics, as a mixin with
    other metrics, or as a composite metric for multiple landfall metrics.

    Public methods:
        maybe_prepare_composite_kwargs: Prepare kwargs for landfall composites
    """

    def __init__(
        self,
        name: str = "landfall_metrics",
        preserve_dims: str = "init_time",
        approach: Literal["first", "next"] = "first",
        exclude_post_landfall: bool = False,
        forecast_variable: Optional[str | derived.DerivedVariable] = None,
        target_variable: Optional[str | derived.DerivedVariable] = None,
        metrics: Optional[list[Type["LandfallMetric"]]] = None,
        *args,
        **kwargs,
    ):
        """Initialize LandfallMetric.

        Landfalls are detected using the calc.find_landfalls function, which utilizes a
        land geometry and line segments based on coordinates to determine intersections.

        Using approach, "first" will calculate the first detected landfall for an entire
        forecast, i.e. later landfalls in a multi-landfall event will not be considered.
        "next" will calculate the next landfall for each init_time.
        Using Ida as an example (case 220), "first" would only run calculations for the
        first landfall in Cuba, ignoring the later US landfall. "next" would run
        calculations for the first landfall in Cuba, then the next landfall in the US,
        etc. based on the init_time and when landfall occurred.

        Args:
            name: The name of the metric. Defaults to "landfall_metrics" for the base
                class.
            preserve_dims: The dimensions to preserve. Defaults to "init_time".
            approach: The approach to use for landfall detection. Defaults to "first".
            exclude_post_landfall: Whether to exclude post-landfall data. Defaults to
                False.
            forecast_variable: The forecast variable to use. Defaults to None.
            target_variable: The target variable to use. Defaults to None.
            metrics: A list of metrics to use as a composite. Defaults to None.
        """
        super().__init__(
            name=name,
            preserve_dims=preserve_dims,
            forecast_variable=forecast_variable,
            target_variable=target_variable,
            *args,
            **kwargs,
        )
        self.approach = approach
        self.exclude_post_landfall = exclude_post_landfall
        self.metrics = metrics or []

        # If metrics provided, instantiate them
        if self.metrics is not None:
            self._metric_instances = [
                (
                    metric_cls(
                        preserve_dims=self.preserve_dims,
                        forecast_variable=self.forecast_variable,
                        target_variable=self.target_variable,
                    )
                    if isinstance(metric_cls, type)
                    else metric_cls
                )
                for metric_cls in self.metrics
            ]
        else:
            self._metric_instances = []

    def __call__(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute the metric.

        Args:
            forecast: The forecast DataArray
            target: The target DataArray
            **kwargs: Additional keyword arguments
        """
        return self.compute_metric(forecast, target, **kwargs)

    def maybe_compute_landfalls(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> tuple[xr.DataArray, xr.DataArray]:
        """Compute landfalls for a given forecast and target dataarray.

        This function computes the landfalls for a given forecast and target dataarray
        using calc.find_landfalls. Currently, this access pattern doesn't include
        passing land geometry in, but calc.find_landfalls will use NaturalEarth's 10m
        land geometry by default.

        Args:
            forecast: The forecast DataArray
            target: The target DataArray
            **kwargs: Additional keyword arguments, may include pre-computed
                forecast_landfall and target_landfall

        Returns:
            Tuple of (forecast_landfall, target_landfall). If no landfalls are found,
            returns NaN DataArrays with init_time dimension.
        """
        forecast_landfall, target_landfall = (
            kwargs.get("forecast_landfall", None),
            kwargs.get("target_landfall", None),
        )
        if forecast_landfall is not None and target_landfall is not None:
            return forecast_landfall, target_landfall

        # For "first" approach: get only first landfall
        # For "next" approach: get all target landfalls, then filter
        return_next_landfall = self.approach == "next"

        # Get first forecast landfall per init_time
        forecast_landfalls = calc.find_landfalls(
            forecast, return_next_landfall=return_next_landfall
        )

        # If no forecast landfalls, return NaN DataArrays for both forecast and target
        if forecast_landfalls is None:
            nan_landfalls = utils._create_nan_dataarray(self.preserve_dims)
            return (nan_landfalls, nan_landfalls.copy())

        # Get only first forecast landfall per init_time
        if "landfall" in forecast_landfalls.dims:
            forecast_landfalls = forecast_landfalls.isel(landfall=0)

        # Get all target landfalls
        target_landfalls_pre_init = calc.find_landfalls(
            target, return_next_landfall=return_next_landfall
        )

        # If no target landfalls, return NaN DataArrays for both forecast and target
        if target_landfalls_pre_init is None:
            nan_landfalls = utils._create_nan_dataarray(self.preserve_dims)
            return (nan_landfalls, nan_landfalls.copy())

        if return_next_landfall:
            # Find next target landfall for each init_time
            target_landfalls = calc.find_next_landfall_for_init_time(
                forecast_landfalls, target_landfalls_pre_init
            )
        else:
            if "landfall" in target_landfalls_pre_init.dims:
                target_landfalls = target_landfalls_pre_init.isel(landfall=0)
            else:
                target_landfalls = target_landfalls_pre_init
            forecast_landfalls = forecast_landfalls.where(
                forecast_landfalls.init_time < target_landfalls.valid_time.values,
                drop=True,
            )
        return forecast_landfalls, target_landfalls

    def maybe_prepare_composite_kwargs(
        self,
        forecast_data: xr.DataArray,
        target_data: xr.DataArray,
        **base_kwargs: Any,
    ) -> dict:
        """Prepare kwargs for composite metric evaluation.

        Computes the landfalls once and adds them to kwargs to avoid recomputing when
        used as a composite metric.

        Args:
            forecast_data: The forecast DataArray.
            target_data: The target DataArray.

        Returns:
            Dictionary of kwargs including transformed_manager.
        """
        kwargs = base_kwargs.copy()

        if self.is_composite() and len(self._metric_instances) > 1:
            kwargs["forecast_landfall"], kwargs["target_landfall"] = (
                self.maybe_compute_landfalls(forecast=forecast_data, target=target_data)
            )

        kwargs["preserve_dims"] = self.preserve_dims

        return kwargs

    def _compute_metric(
        self,
        forecast: xr.DataArray,
        target: xr.DataArray,
        **kwargs: Any,
    ) -> Any:
        """Compute metric (not supported for LandfallMetric base)

        LandfallMetric must be subclassed (like LandfallDisplacement,
        LandfallTimeMeanError)
        or used as a composite with metrics list.

        Args:
            forecast: The forecast DataArray
            target: The target DataArray
            **kwargs: Additional keyword arguments
        """
        raise NotImplementedError(
            "LandfallMetric._compute_metric must be implemented "
            "by subclasses (LandfallDisplacement, LandfallTimeMeanError, etc.) or use "
            "LandfallMetric as a composite with metrics=[...] list. Composites are "
            "automatically expanded in the evaluation pipeline."
        )


class SpatialDisplacement(BaseMetric):
    """Compute spatial displacement between forecast and target patterns.

    Extends BaseMetric to compute great circle distance between centers of
    mass of forecast and target spatial patterns. Useful for atmospheric
    rivers and similar spatial features.
    """

    def __init__(
        self,
        name: str = "spatial_displacement",
        **kwargs: Any,
    ):
        """Initialize the Spatial Displacement metric.

        Args:
            name: The name of the metric. Defaults to
                "spatial_displacement".
            **kwargs: Additional keyword arguments passed to BaseMetric.
        """
        super().__init__(name, **kwargs)

    def _compute_metric(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute spatial displacement.

        Args:
            forecast: The forecast DataArray.
            target: The target DataArray.

        Returns:
            The spatial displacement between the forecast and target as a DataArray.
        """

        def center_of_mass_ufunc(data):
            """ufunc tooling to calculate the center of mass of a 2D array, returning
            a tuple of the latitude and longitude indices, or np.nan tuple if no
            non-zero values are present.
            """
            if (data > 0).any():
                return ndimage.center_of_mass(data)
            else:
                return (np.nan, np.nan)

        target_lat_idx, target_lon_idx = xr.apply_ufunc(
            center_of_mass_ufunc,
            target,
            input_core_dims=[["latitude", "longitude"]],
            output_core_dims=[[], []],
            vectorize=True,
            dask="allowed",
        )

        # Process target coordinates
        target_lat_idx = np.round(target_lat_idx)
        target_lon_idx = np.round(target_lon_idx)
        target_lat_coords, target_lon_coords = utils.idx_to_coords(
            target_lat_idx,
            target_lon_idx,
            target.latitude.values,
            target.longitude.values,
        )
        target_coordinates = np.array([target_lat_coords, target_lon_coords])

        # Process forecast coordinates
        forecast_lat_idx, forecast_lon_idx = xr.apply_ufunc(
            center_of_mass_ufunc,
            forecast,
            input_core_dims=[["latitude", "longitude"]],
            output_core_dims=[[], []],
            vectorize=True,
            dask="allowed",
        )
        forecast_lat_idx = np.round(forecast_lat_idx)
        forecast_lon_idx = np.round(forecast_lon_idx)
        forecast_lat_coords, forecast_lon_coords = utils.idx_to_coords(
            forecast_lat_idx,
            forecast_lon_idx,
            forecast.latitude.values,
            forecast.longitude.values,
        )
        forecast_coordinates = np.array([forecast_lat_coords, forecast_lon_coords])

        # Calculate haversine distance
        distance = calc.haversine_distance(forecast_coordinates, target_coordinates)

        # Create DataArray with all dimensions
        result = xr.DataArray(
            distance,
            coords={"lead_time": forecast.lead_time, "valid_time": forecast.valid_time},
            dims=["lead_time", "valid_time"],
            name="spatial_displacement",
        )

        # Reduce over non-preserved dimensions (valid_time) by taking mean
        time_dims_to_reduce = [
            dim for dim in result.dims if dim not in self.preserve_dims
        ]
        if time_dims_to_reduce:
            result = result.mean(dim=time_dims_to_reduce)

        return result


class LandfallDisplacement(LandfallMetric):
    """Compute distance between forecast and target landfall positions.

    Extends LandfallMetric to calculate the spatial distance between forecast
    and target landfall positions, defaulting to kilometers.
    """

    def __init__(
        self,
        name: str = "landfall_displacement",
        *args,
        **kwargs,
    ):
        """Initialize the Landfall Displacement metric.

        Args:
            name: The name of the metric. Defaults to
                "landfall_displacement".
            *args: Additional positional arguments passed to LandfallMetric.
            **kwargs: Additional keyword arguments passed to LandfallMetric.
        """
        super().__init__(name, *args, **kwargs)
        self.units = kwargs.get("units", "km")

    def calculate_displacement(
        self,
        forecast_landfall: xr.DataArray,
        target_landfall: xr.DataArray,
        units: Literal["km", "kilometers", "deg", "degrees"] = "km",
    ) -> xr.DataArray:
        """Calculate the distance between two landfall points in kilometers or degrees.

        Handles both scalar and multi-dimensional (with init_time) DataArrays.

        Args:
            forecast_landfall: Forecast landfall xarray DataArray
            target_landfall: Target landfall xarray DataArray
            units: The units to use for the distance. Defaults to "km"
        Returns:
            Distance in the specified units as xarray DataArray
        """
        # Find common init_times between forecast and target
        common_init_times = utils.find_common_init_times(
            forecast_landfall, target_landfall
        )

        # If no common init_times, return NaN DataArray
        if not common_init_times:
            return utils._create_nan_dataarray(self.preserve_dims)

        # Compute distance for each common init_time
        distances = []
        for init_time in common_init_times:
            f_lat = forecast_landfall.sel(init_time=init_time).coords["latitude"].values
            f_lon = (
                forecast_landfall.sel(init_time=init_time).coords["longitude"].values
            )
            t_lat = target_landfall.sel(init_time=init_time).coords["latitude"].values
            t_lon = target_landfall.sel(init_time=init_time).coords["longitude"].values

            # Skip if any coordinates are NaN
            if (
                np.any(np.isnan(f_lat))
                or np.any(np.isnan(f_lon))
                or np.any(np.isnan(t_lat))
                or np.any(np.isnan(t_lon))
            ):
                distances.append(np.nan)
            else:
                dist = calc.haversine_distance(
                    [f_lat, f_lon], [t_lat, t_lon], units=units
                )
                # Ensure we append a scalar value
                distances.append(
                    float(dist.item()) if hasattr(dist, "item") else float(dist)
                )

        return xr.DataArray(
            distances,
            dims=["init_time"],
            coords={"init_time": common_init_times},
        )

    def _compute_metric(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute the landfall displacement metric."""
        forecast_landfall, target_landfall = self.maybe_compute_landfalls(
            forecast, target, **kwargs
        )
        if not utils.is_valid_landfall(
            forecast_landfall
        ) or not utils.is_valid_landfall(target_landfall):
            return utils._create_nan_dataarray(self.preserve_dims)
        return self.calculate_displacement(
            forecast_landfall,
            target_landfall,
            units=self.units,
        )


class LandfallTimeMeanError(LandfallMetric):
    """Compute mean error between forecast and target landfall times.

    Extends LandfallMetric to calculate timing difference. Positive values
    indicate forecast landfall is later than target; negative values indicate
    forecast landfall is earlier than target.
    """

    def __init__(
        self,
        name: str = "landfall_time_me",
        *args,
        **kwargs,
    ):
        """Initialize the Landfall Time Mean Error metric.

        Args:
            name: The name of the metric. Defaults to "landfall_time_me".
            *args: Additional positional arguments passed to LandfallMetric.
            **kwargs: Additional keyword arguments passed to LandfallMetric.
        """
        super().__init__(name, *args, **kwargs)

    def calculate_time_difference(
        self,
        forecast_landfall: xr.DataArray,
        target_landfall: xr.DataArray,
    ) -> xr.DataArray:
        """Calculate the time difference between two landfall points in hours.

        Args:
            forecast_landfall: Forecast landfall xarray DataArray.
            target_landfall: Target landfall xarray DataArray.

        Returns:
            Time difference in hours (forecast_landfall - target_landfall)
            as xarray DataArray with init_time dimension.
        """
        # Find common init_times between forecast and target
        common_init_times = utils.find_common_init_times(
            forecast_landfall, target_landfall
        )

        # If no common init_times, return NaN DataArray
        if not common_init_times:
            return utils._create_nan_dataarray(self.preserve_dims)

        # Calculate time difference for each common init_time
        time_diffs = []
        for init_time in common_init_times:
            time1 = forecast_landfall.sel(init_time=init_time).coords["valid_time"]
            time2 = target_landfall.sel(init_time=init_time).coords["valid_time"]

            # Calculate time difference in hours
            time_diff = (time1 - time2) / np.timedelta64(1, "h")
            time_diffs.append(float(time_diff.values))

        return xr.DataArray(
            time_diffs,
            dims=self.preserve_dims,
            coords={self.preserve_dims: common_init_times},
        )

    def _compute_metric(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute the landfall time metric."""
        forecast_landfall, target_landfall = self.maybe_compute_landfalls(
            forecast, target, **kwargs
        )
        if not utils.is_valid_landfall(
            forecast_landfall
        ) or not utils.is_valid_landfall(target_landfall):
            return utils._create_nan_dataarray(self.preserve_dims)
        return self.calculate_time_difference(forecast_landfall, target_landfall)


class LandfallIntensityMeanAbsoluteError(LandfallMetric, MeanAbsoluteError):
    """Compute MAE of forecast and target intensity at landfall.

    Extends both LandfallMetric and MeanAbsoluteError to calculate mean
    absolute error between forecast and target intensity at landfall time.

    The intensity variable is determined by forecast_variable and
    target_variable. For multiple intensity variables, create separate metric
    instances for each variable.
    """

    def __init__(
        self,
        name: str = "landfall_intensity_mae",
        *args,
        **kwargs,
    ):
        """Initialize the Landfall Intensity Mean Absolute Error metric.

        Args:
            name: The name of the metric. Defaults to
                "landfall_intensity_mae".
            *args: Additional positional arguments passed to parent classes.
            **kwargs: Additional keyword arguments passed to parent classes.
        """
        super().__init__(name, *args, **kwargs)

    def _compute_metric(
        self, forecast: xr.DataArray, target: xr.DataArray, **kwargs: Any
    ) -> Any:
        """Compute the landfall intensity metric."""
        forecast_landfall, target_landfall = self.maybe_compute_landfalls(
            forecast, target, **kwargs
        )
        if not utils.is_valid_landfall(
            forecast_landfall
        ) or not utils.is_valid_landfall(target_landfall):
            return utils._create_nan_dataarray(self.preserve_dims)

        # The complexity of the landfall outputs makes it easier just to use np.abs here
        return np.abs(forecast_landfall - target_landfall)



import importlib.util
import os
import pathlib
import pickle
from typing import Optional

import click
import pandas as pd

import extremeweatherbench.cases as cases
import extremeweatherbench.defaults as defaults
import extremeweatherbench.evaluate as evaluate


@click.command()
@click.option(
    "--default",
    is_flag=True,
    help="Use default Brightband evaluation objects with current directory as output",
)
@click.option(
    "--config-file",
    type=click.Path(exists=True),
    help="Path to a config.py file containing evaluation objects",
)
@click.option(
    "--output-dir",
    type=click.Path(),
    help="Directory for analysis outputs (default: current directory)",
)
@click.option(
    "--cache-dir",
    type=click.Path(),
    help="Optional directory for caching intermediate data",
)
@click.option(
    "--n-jobs",
    type=int,
    default=1,
    help="Number of parallel jobs to run (default: 1 for serial execution)",
)
@click.option(
    "--parallel-config",
    "-p",
    type=dict,
    default=None,
    help=(
        "Advanced parallel configuration using joblib. Takes precedence over "
        "--n-jobs if provided."
    ),
)
@click.option(
    "--save-case-operators",
    type=click.Path(),
    help="Save CaseOperator objects to a pickle file at this path",
)
@click.pass_context
def cli_runner(
    ctx: click.Context,
    default: bool,
    config_file: Optional[str],
    output_dir: Optional[str],
    cache_dir: Optional[str],
    n_jobs: int,
    parallel_config: Optional[dict],
    save_case_operators: Optional[str],
):
    """ExtremeWeatherBench command line interface.

    This CLI supports two main modes:

    1. Default mode (--default): Uses the predefined Brightband evaluation objects for
       comprehensive weather event evaluation including heat waves, freeze events,
       [severe convection, atmospheric rivers, and tropical cyclones] (bracketed events
       are not yet implemented).

    2. Custom mode (--config-file): Uses a Python config file containing custom
       evaluation objects defined by the user.

    The CLI can run evaluations in serial or parallel using joblib, and optionally
    save CaseOperator objects for later use or inspection.

    Args:
        default: Use default Brightband evaluation objects with current directory
            as output.
        config_file: Path to a config.py file containing evaluation objects.
        output_dir: Directory for analysis outputs (default: current directory).
        cache_dir: Optional directory for caching intermediate data. When set,
            datasets or dataarrays are computed and cached as zarrs.
        n_jobs: Number of parallel jobs to run (default: 1 for serial execution).
        parallel_config: Advanced parallel configuration using joblib. Takes
            precedence over n_jobs if provided.
        save_case_operators: Save CaseOperator objects to a pickle file at this
            path.
    Examples:
        # Use default evaluation objects
        $ ewb --default

        # Use custom config file with parallel execution
        $ ewb --config-file my_config.py --n-jobs 4

        # Save case operators to pickle file
        $ ewb --default --save-case-operators case_ops.pkl

        # Use custom output and cache directories (cache enables zarr storage)
        $ ewb --default --output-dir ./results --cache-dir ./cache

        # Use custom parallel configuration
        $ ewb --default --parallel-config '{"backend": "dask", "n_jobs": 4}'
    """
    # Show help if no arguments provided
    if not default and not config_file:
        click.echo(ctx.get_help())
        ctx.exit()

    # Set default output directory to current working directory
    if output_dir is None:
        output_dir = os.getcwd()

    output_path = pathlib.Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)

    if default and config_file:
        raise click.UsageError("Cannot specify both --default and --config-file")

    # Load evaluation objects
    if default:
        click.echo("Using default Brightband evaluation objects...")
        evaluation_objects = defaults.get_brightband_evaluation_objects()
        case_list = _load_default_cases()
    else:
        assert config_file is not None  # for mypy
        click.echo(f"Loading evaluation objects from {config_file}...")
        evaluation_objects, case_list = _load_config_file(config_file)

    # Initialize ExtremeWeatherBench
    ewb = evaluate.ExtremeWeatherBench(
        case_metadata=case_list,
        evaluation_objects=evaluation_objects,
        cache_dir=cache_dir if cache_dir else None,
    )

    # Get case operators
    case_operators = ewb.case_operators
    click.echo(f"Found {len(case_operators)} case operators to evaluate")

    # Save case operators if requested
    if save_case_operators:
        save_path = pathlib.Path(save_case_operators)
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with open(save_path, "wb") as f:
            pickle.dump(case_operators, f)
        click.echo(f"Case operators saved to {save_case_operators}")

    # Run evaluation
    click.echo("Running evaluation...")
    results = ewb.run_evaluation(
        n_jobs=n_jobs,
        parallel_config=parallel_config,
    )

    # Save results
    output_file = output_path / "evaluation_results.csv"
    if isinstance(results, pd.DataFrame) and not results.empty:
        results.to_csv(output_file, index=False)
        click.echo(f"Results saved to {output_file}")
        click.echo(f"Evaluated {len(results)} cases")
    else:
        click.echo("No results to save")


def _load_default_cases():
    """Load default case data for default evaluation objects."""

    return cases.load_ewb_events_yaml_into_case_list()


def _load_config_file(config_path: str) -> tuple:
    """Load evaluation objects and cases from a Python config file.

    The config file should define:
    - evaluation_objects: List of EvaluationObject instances
    - case_list: List of case data
    """
    config_path_obj = pathlib.Path(config_path)

    # Load the config module
    spec = importlib.util.spec_from_file_location("config", str(config_path_obj))
    if spec is None or spec.loader is None:
        raise click.ClickException(f"Could not load config file: {config_path}")

    config_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config_module)

    # Extract required attributes
    if not hasattr(config_module, "evaluation_objects"):
        raise click.ClickException("Config file must define 'evaluation_objects' list")

    if not hasattr(config_module, "case_list"):
        raise click.ClickException("Config file must define 'case_list' list")

    return config_module.evaluation_objects, config_module.case_list


if __name__ == "__main__":
    cli_runner()



"""ExtremeWeatherBench: A benchmarking framework for extreme weather forecasts.

This module provides the public API for ExtremeWeatherBench. Users can import
the package and access all key functionality:

    import extremeweatherbench as ewb

    # Main entry point for evaluation
    ewb.evaluation(case_metadata=..., evaluation_objects=...)

    # Hierarchical access via namespace submodules
    ewb.targets.ERA5(...)
    ewb.forecasts.ZarrForecast(...)
    ewb.metrics.MeanAbsoluteError(...)

    # Also available at top level
    ewb.ERA5(...)
    ewb.load_cases()
"""

from types import SimpleNamespace

# Import actual modules for backwards compatibility
from extremeweatherbench import calc, cases, defaults, derived, metrics, regions, utils

# Import specific items for top-level access
from extremeweatherbench.calc import (
    convert_from_cartesian_to_latlon,
    geopotential_thickness,
    great_circle_mask,
    haversine_distance,
    maybe_calculate_wind_speed,
    mixing_ratio,
    orography,
    pressure_at_surface,
    saturation_mixing_ratio,
    saturation_vapor_pressure,
    specific_humidity_from_relative_humidity,
)
from extremeweatherbench.cases import (
    CaseOperator,
    IndividualCase,
    build_case_operators,
    load_ewb_events_yaml_into_case_list,
    load_individual_cases,
    load_individual_cases_from_yaml,
    read_incoming_yaml,
)
from extremeweatherbench.defaults import (
    DEFAULT_COORDINATE_VARIABLES,
    DEFAULT_VARIABLE_NAMES,
    cira_fcnv2_atmospheric_river_forecast,
    cira_fcnv2_freeze_forecast,
    cira_fcnv2_heatwave_forecast,
    cira_fcnv2_severe_convection_forecast,
    cira_fcnv2_tropical_cyclone_forecast,
    era5_atmospheric_river_target,
    era5_freeze_target,
    era5_heatwave_target,
    get_brightband_evaluation_objects,
    get_climatology,
    ghcn_freeze_target,
    ghcn_heatwave_target,
    ibtracs_target,
    lsr_target,
    pph_target,
)
from extremeweatherbench.derived import (
    AtmosphericRiverVariables,
    CravenBrooksSignificantSevere,
    DerivedVariable,
    TropicalCycloneTrackVariables,
    maybe_derive_variables,
    maybe_include_variables_from_derived_input,
)
from extremeweatherbench.evaluate import ExtremeWeatherBench
from extremeweatherbench.inputs import (
    ARCO_ERA5_FULL_URI,
    DEFAULT_GHCN_URI,
    ERA5,
    GHCN,
    IBTRACS_URI,
    LSR,
    LSR_URI,
    PPH,
    PPH_URI,
    CIRA_metadata_variable_mapping,
    ERA5_metadata_variable_mapping,
    EvaluationObject,
    ForecastBase,
    HRES_metadata_variable_mapping,
    IBTrACS,
    IBTrACS_metadata_variable_mapping,
    InputBase,
    KerchunkForecast,
    TargetBase,
    XarrayForecast,
    ZarrForecast,
    align_forecast_to_target,
    check_for_missing_data,
    maybe_subset_variables,
    open_kerchunk_reference,
    zarr_target_subsetter,
)
from extremeweatherbench.metrics import (
    Accuracy,
    BaseMetric,
    CompositeMetric,
    CriticalSuccessIndex,
    DurationMeanError,
    EarlySignal,
    FalseAlarmRatio,
    FalseNegatives,
    FalsePositives,
    LandfallDisplacement,
    LandfallIntensityMeanAbsoluteError,
    LandfallMetric,
    LandfallTimeMeanError,
    MaximumLowestMeanAbsoluteError,
    MaximumMeanAbsoluteError,
    MeanAbsoluteError,
    MeanError,
    MeanSquaredError,
    MinimumMeanAbsoluteError,
    RootMeanSquaredError,
    SpatialDisplacement,
    ThresholdMetric,
    TrueNegatives,
    TruePositives,
)
from extremeweatherbench.regions import (
    REGION_TYPES,
    BoundingBoxRegion,
    CenteredRegion,
    Region,
    RegionSubsetter,
    ShapefileRegion,
    map_to_create_region,
    subset_cases_to_region,
    subset_results_to_region,
)
from extremeweatherbench.utils import (
    check_for_vars,
    convert_day_yearofday_to_time,
    convert_init_time_to_valid_time,
    convert_longitude_to_180,
    convert_longitude_to_360,
    convert_valid_time_to_init_time,
    derive_indices_from_init_time_and_lead_time,
    determine_temporal_resolution,
    extract_tc_names,
    filter_kwargs_for_callable,
    find_common_init_times,
    idx_to_coords,
    interp_climatology_to_target,
    is_valid_landfall,
    load_land_geometry,
    maybe_cache_and_compute,
    maybe_densify_dataarray,
    maybe_get_closest_timestamp_to_center_of_valid_times,
    maybe_get_operator,
    min_if_all_timesteps_present,
    min_if_all_timesteps_present_forecast,
    read_event_yaml,
    remove_ocean_gridpoints,
    stack_dataarray_from_dims,
)

# Aliases
evaluation = ExtremeWeatherBench
load_cases = load_ewb_events_yaml_into_case_list

# Namespace submodules for convenient grouping (these don't shadow actual modules)
targets = SimpleNamespace(
    TargetBase=TargetBase,
    ERA5=ERA5,
    GHCN=GHCN,
    IBTrACS=IBTrACS,
    LSR=LSR,
    PPH=PPH,
)

forecasts = SimpleNamespace(
    ForecastBase=ForecastBase,
    ZarrForecast=ZarrForecast,
    KerchunkForecast=KerchunkForecast,
    XarrayForecast=XarrayForecast,
)

__all__ = [
    # Core evaluation
    "evaluation",
    "ExtremeWeatherBench",
    # Modules
    "calc",
    "cases",
    "defaults",
    "derived",
    "metrics",
    "regions",
    "utils",
    # Namespace submodules
    "targets",
    "forecasts",
    # Aliases
    "load_cases",
    # calc
    "convert_from_cartesian_to_latlon",
    "geopotential_thickness",
    "great_circle_mask",
    "haversine_distance",
    "maybe_calculate_wind_speed",
    "mixing_ratio",
    "orography",
    "pressure_at_surface",
    "saturation_mixing_ratio",
    "saturation_vapor_pressure",
    "specific_humidity_from_relative_humidity",
    # cases
    "CaseOperator",
    "IndividualCase",
    "build_case_operators",
    "load_ewb_events_yaml_into_case_list",
    "load_individual_cases",
    "load_individual_cases_from_yaml",
    "read_incoming_yaml",
    # defaults
    "DEFAULT_COORDINATE_VARIABLES",
    "DEFAULT_VARIABLE_NAMES",
    "cira_fcnv2_atmospheric_river_forecast",
    "cira_fcnv2_freeze_forecast",
    "cira_fcnv2_heatwave_forecast",
    "cira_fcnv2_severe_convection_forecast",
    "cira_fcnv2_tropical_cyclone_forecast",
    "era5_atmospheric_river_target",
    "era5_freeze_target",
    "era5_heatwave_target",
    "get_brightband_evaluation_objects",
    "get_climatology",
    "ghcn_freeze_target",
    "ghcn_heatwave_target",
    "ibtracs_target",
    "lsr_target",
    "pph_target",
    # derived
    "AtmosphericRiverVariables",
    "CravenBrooksSignificantSevere",
    "DerivedVariable",
    "TropicalCycloneTrackVariables",
    "maybe_derive_variables",
    "maybe_include_variables_from_derived_input",
    # inputs
    "ARCO_ERA5_FULL_URI",
    "CIRA_metadata_variable_mapping",
    "DEFAULT_GHCN_URI",
    "ERA5",
    "ERA5_metadata_variable_mapping",
    "EvaluationObject",
    "ForecastBase",
    "GHCN",
    "HRES_metadata_variable_mapping",
    "IBTrACS",
    "IBTrACS_metadata_variable_mapping",
    "IBTRACS_URI",
    "InputBase",
    "KerchunkForecast",
    "LSR",
    "LSR_URI",
    "PPH",
    "PPH_URI",
    "TargetBase",
    "XarrayForecast",
    "ZarrForecast",
    "align_forecast_to_target",
    "check_for_missing_data",
    "maybe_subset_variables",
    "open_kerchunk_reference",
    "zarr_target_subsetter",
    # metrics
    "Accuracy",
    "BaseMetric",
    "CompositeMetric",
    "CriticalSuccessIndex",
    "DurationMeanError",
    "EarlySignal",
    "FalseAlarmRatio",
    "FalseNegatives",
    "FalsePositives",
    "LandfallDisplacement",
    "LandfallIntensityMeanAbsoluteError",
    "LandfallMetric",
    "LandfallTimeMeanError",
    "MaximumLowestMeanAbsoluteError",
    "MaximumMeanAbsoluteError",
    "MeanAbsoluteError",
    "MeanError",
    "MeanSquaredError",
    "MinimumMeanAbsoluteError",
    "RootMeanSquaredError",
    "SpatialDisplacement",
    "ThresholdMetric",
    "TrueNegatives",
    "TruePositives",
    # regions
    "BoundingBoxRegion",
    "CenteredRegion",
    "REGION_TYPES",
    "Region",
    "RegionSubsetter",
    "ShapefileRegion",
    "map_to_create_region",
    "subset_cases_to_region",
    "subset_results_to_region",
    # utils
    "check_for_vars",
    "convert_day_yearofday_to_time",
    "convert_init_time_to_valid_time",
    "convert_longitude_to_180",
    "convert_longitude_to_360",
    "convert_valid_time_to_init_time",
    "derive_indices_from_init_time_and_lead_time",
    "determine_temporal_resolution",
    "extract_tc_names",
    "filter_kwargs_for_callable",
    "find_common_init_times",
    "idx_to_coords",
    "interp_climatology_to_target",
    "is_valid_landfall",
    "load_land_geometry",
    "maybe_cache_and_compute",
    "maybe_densify_dataarray",
    "maybe_get_closest_timestamp_to_center_of_valid_times",
    "maybe_get_operator",
    "min_if_all_timesteps_present",
    "min_if_all_timesteps_present_forecast",
    "read_event_yaml",
    "remove_ocean_gridpoints",
    "stack_dataarray_from_dims",
]



"""Input source modules for different data types."""

import pandas as pd
import polars as pl
import xarray as xr

from . import pandas_dataframe, polars_lazyframe, xarray_dataarray, xarray_dataset
from .base import Source

# Registry mapping data types to their corresponding source modules
# Each module implements the Source Protocol at module level
DATA_BACKEND_REGISTRY: dict[type, Source] = {
    pd.DataFrame: pandas_dataframe,
    pl.LazyFrame: polars_lazyframe,
    xr.DataArray: xarray_dataarray,
    xr.Dataset: xarray_dataset,
}


def get_backend_module(data_type: type) -> Source:
    """Get the source module for a given data type.

    Args:
        data_type: The type of data (e.g., pd.DataFrame, xr.Dataset)

    Returns:
        The module that implements Source Protocol for this data type

    Raises:
        ValueError: If no source handler is registered for the data type
    """
    module = DATA_BACKEND_REGISTRY.get(data_type)
    if module is None:
        available_types = list(DATA_BACKEND_REGISTRY.keys())
        raise ValueError(
            f"No source handler found for dataset type: {data_type}. "
            f"Available types: {available_types}"
        )

    return module



from typing import Literal, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
import regionmask
import scores.categorical as categorical
import shapely
import xarray as xr

from extremeweatherbench import utils

epsilon: float = 0.6219569100577033  # Ratio of molecular weights (H2O/dry air)
sat_press_0c: float = 6.112  # Saturation vapor pressure at 0°C (hPa)
g0: float = 9.80665  # Standard gravity (m/s^2)


def convert_from_cartesian_to_latlon(
    input_point: Union[np.ndarray, tuple[float, float]],
    latitude: xr.DataArray,
    longitude: xr.DataArray,
) -> tuple[float, float]:
    """Convert point from cartesian coordinate system to lat/lon.

    Args:
        input_point: Point as tuple (y, x) in cartesian system
        latitude: Latitude DataArray
        longitude: Longitude DataArray

    Returns:
        Tuple (latitude, longitude) in degrees
    """
    lat_idx = int(input_point[0])
    lon_idx = int(input_point[1])
    return (
        latitude.isel(latitude=lat_idx).values,
        longitude.isel(longitude=lon_idx).values,
    )


def mixing_ratio(
    partial_pressure: float | npt.NDArray, total_pressure: float | npt.NDArray
) -> float | npt.NDArray[np.float64]:
    r"""Calculate the mixing ratio of water vapor in air.

    Uses the formula: $w = (\epsilon * e) / (p - e)$ where $\epsilon = 0.622$.

    Args:
        partial_pressure: Water vapor partial pressure in hPa.
        total_pressure: Total atmospheric pressure in hPa.

    Returns:
        numpy.ndarray: Mixing ratio in kg/kg (dimensionless).

    Notes:
        - Mixing ratio is approximately constant with height for unsaturated air
        - Values typically range from 0 to ~0.025 kg/kg in the atmosphere
        - ε (epsilon) = 0.622 is the ratio of molecular weights (H2O/dry air)
    """
    # Suppress warnings for this specific calculation
    with np.errstate(divide="ignore", invalid="ignore"):
        return epsilon * partial_pressure / (total_pressure - partial_pressure)


def saturation_vapor_pressure(
    temperature: float | npt.NDArray,
) -> npt.NDArray[np.float64]:
    r"""Calculate saturation vapor pressure using the Clausius-Clapeyron equation.

    Uses the Magnus formula approximation which is accurate for temperatures
    between -40°C and +50°C. Formula:

    $$e_ = 6.112 \times \exp\left(\frac{17.67*T}{T+243.5}\right)$$

    Args:
        temperature: Temperature values in Celsius (can be scalar or array).

    Returns:
        numpy.ndarray: Saturation vapor pressure values in hPa.

    Notes:
        - Based on the Magnus formula which is accurate within ±0.1% for typical
          atmospheric temperatures
        - Saturation vapor pressure increases exponentially with temperature
        - At 0°C: ~6.11 hPa, at 20°C: ~23.4 hPa, at 30°C: ~42.4 hPa
    """
    # Suppress overflow warnings for this calculation
    with np.errstate(over="ignore", invalid="ignore"):
        return sat_press_0c * np.exp(17.67 * temperature / (temperature + 243.5))


def saturation_mixing_ratio(
    pressure: float | npt.NDArray, temperature: float | npt.NDArray
) -> npt.NDArray[np.float64]:
    """Calculates the saturation mixing ratio of a parcel.

    Args:
        pressure: Pressure values in hPa
        temperature: Temperature values in C

    Returns:
        Saturation mixing ratio values in kg/kg
    """
    return mixing_ratio(saturation_vapor_pressure(temperature), pressure)


def haversine_distance(
    input_a: Sequence[Union[float, xr.DataArray]],
    input_b: Sequence[Union[float, xr.DataArray]],
    units: Literal["km", "kilometers", "deg", "degrees"] = "km",
) -> Union[float, xr.DataArray]:
    """Calculate the great-circle/haversine distance between two points on the Earth's
    surface.

    Args:
        input_a: The first point, represented as an ndarray of shape (2,n) in
            degrees lat/lon.
        input_b: The second point(s), represented as an ndarray of shape (2,n)
            in degrees lat/lon.

    Returns:
        The great-circle distance between the two points in kilometers.
    """
    # Convert to radians for calculations
    lat1 = np.radians(input_a[0])
    lon1 = np.radians(input_a[1])
    lat2 = np.radians(input_b[0])
    lon2 = np.radians(input_b[1])

    # Haversine formula for great circle distance
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    c = 2 * np.arcsin(np.sqrt(a))
    if units == "km" or units == "kilometers":
        return 6371 * c
    elif units == "deg" or units == "degrees":
        return np.degrees(c)  # Convert back to degrees
    else:
        raise ValueError(f"Invalid units: {units}")


def great_circle_mask(
    ds: xr.Dataset, latlon_point: tuple[float, float], radius_degrees: float
) -> xr.DataArray:
    """Create a circular mask based on great circle distance for an xarray dataset.

    Args:
        ds: Dataset with 'latitude' and 'longitude' coordinates.
        latlon_point: Tuple containing (latitude, longitude) of the center point.
        radius_degrees: Radius in degrees of great circle distance.

    Returns:
        Boolean mask where True indicates points within the radius.
    """

    distance = haversine_distance(
        latlon_point, (ds.latitude, ds.longitude), units="deg"
    )
    # Create mask as xarray DataArray
    if isinstance(distance, xr.DataArray):
        mask = distance <= radius_degrees
    else:
        # If distance is a scalar, create a DataArray mask
        mask = xr.full_like(ds.latitude, distance <= radius_degrees, dtype=bool)

    return mask


def orography(ds: xr.Dataset) -> xr.DataArray:
    """Calculate the orography from the geopotential at the surface using ERA5.

    Args:
        ds: The potential xarray dataset to calculate the orography from.

    Returns:
        The orography as an xarray DataArray.
    """

    if "geopotential_at_surface" in ds.variables:
        return ds["geopotential_at_surface"].isel(time=0) / g0
    else:
        # Import inputs here to avoid circular import
        from extremeweatherbench import inputs

        era5 = xr.open_zarr(
            inputs.ARCO_ERA5_FULL_URI,
            chunks=None,
            storage_options=dict(token="anon"),
        )
        return (
            era5.isel(time=1000000)["geopotential_at_surface"].sel(
                latitude=ds.latitude, longitude=ds.longitude
            )
            / g0
        )


def pressure_at_surface(orography_da: xr.DataArray) -> xr.DataArray:
    """Calculate the pressure at the surface, based on orography.

    The dataarray is orography (geopotential at the surface/g0).

    Args:
        orography_da: The orography dataarray.

    Returns:
        The pressure at the surface of the dataarray in Pa.
    """
    return 101325 * (1 - 2.25577e-5 * orography_da) ** 5.25579


def maybe_calculate_wind_speed(ds: xr.Dataset) -> xr.Dataset:
    """Maybe prepare wind data by computing wind speed.

    If the wind speed is not already present, it will be computed from the eastward
    and northward wind components (u and v). If the wind speed is already present, the
    dataset is returned as is.

    Args:
        ds: The dataset to prepare the wind data from.

    Returns:
        A dataset with the wind speed computed if it is not already present.
    """

    has_wind_speed = "surface_wind_speed" in ds.data_vars
    has_wind_components = (
        "surface_eastward_wind" in ds.data_vars
        and "surface_northward_wind" in ds.data_vars
    )

    # If we don't have wind speed but have components, compute it
    if not has_wind_speed and has_wind_components:
        ds["surface_wind_speed"] = np.hypot(
            ds["surface_eastward_wind"], ds["surface_northward_wind"]
        )

    return ds


def geopotential_thickness(
    da: xr.DataArray,
    top_level: int = 300,
    bottom_level: int = 500,
    geopotential: bool = False,
    pressure_dim: str = "level",
) -> xr.DataArray:
    """Generate the geopotential thickness from the geopotential heights. Defaults to
    geopotential height in meters.

    Args:
        da: The xarray DataArray to generate the geopotential thickness from.
        top_level: The pressure value of the top level to generate the
            geopotential thickness from in hPa.
        bottom_level: The pressure value of the bottom level to generate the
            geopotential thickness from in hPa.
        geopotential: Whether the input DataArray is geopotential height (in m) or
            geopotential (in m^2/s^2). Default is geopotential height.
        pressure_dim: The name of the pressure dimension. Default is "level".

    Returns:
        The geopotential thickness in meters as an xarray DataArray.
    """
    geopotential_heights = da.sel({pressure_dim: top_level})
    geopotential_height_bottom = da.sel({pressure_dim: bottom_level})
    if geopotential:
        geopotential_thickness = (
            geopotential_heights - geopotential_height_bottom
        ) / g0
    else:
        geopotential_thickness = geopotential_heights - geopotential_height_bottom
    geopotential_thickness.attrs = dict(
        description=f"Geopotential thickness of {top_level} hPa and {bottom_level} hPa",
        units="m",
    )
    return geopotential_thickness


def nantrapezoid(
    y: np.ndarray,
    x: np.ndarray | None = None,
    dx: float = 1.0,
    axis: int = -1,
):
    """Trapezoid rule for arrays with nans.

    Identical to np.trapezoid but with nans handled correctly in the summation.
    """
    y = np.asanyarray(y)
    if x is None:
        # Create an array of the step size
        d = np.full(y.shape[axis] - 1, dx) if y.shape[axis] > 1 else np.array([dx])
        # reshape to correct shape
        shape = [1] * y.ndim
        shape[axis] = d.shape[0]
        d = d.reshape(shape)
    else:
        x = np.asanyarray(x)
        if x.ndim == 1:
            d = np.diff(x)
            # reshape to correct shape
            shape = [1] * y.ndim
            shape[axis] = d.shape[0]
            d = d.reshape(shape)
        else:
            d = np.diff(x, axis=axis)
    if y.ndim != d.ndim:
        d = np.expand_dims(d, axis=1)
    nd = y.ndim
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)
    try:
        # This is the only location different from np.trapezoid
        ret = np.nansum(d * (y[tuple(slice1)] + y[tuple(slice2)]) / 2.0, axis=axis)
    except ValueError:
        # Operations didn't work, cast to ndarray
        d = np.asarray(d)
        y = np.asarray(y)
        ret = np.add.reduce(d * (y[tuple(slice1)] + y[tuple(slice2)]) / 2.0, axis)
    return ret


def specific_humidity_from_relative_humidity(
    air_temperature: xr.DataArray, relative_humidity: xr.DataArray, levels: xr.DataArray
) -> xr.DataArray:
    """Compute specific humidity from relative humidity and air temperature.

    Args:
        air_temperature: Air temperature in Kelvin.
        relative_humidity: Relative humidity (0-1 or 0-100 depending on data).
        levels: Pressure levels in hPa.

    Returns:
        A DataArray of specific humidity in kg/kg.
    """
    # Compute saturation mixing ratio; air temperature must be in Kelvin
    sat_mixing_ratio = saturation_mixing_ratio(levels, air_temperature - 273.15)

    # Calculate specific humidity using saturation mixing ratio, epsilon,
    # and relative humidity
    mixing_ratio = (
        epsilon
        * sat_mixing_ratio
        * relative_humidity
        / (epsilon + sat_mixing_ratio * (1 - relative_humidity))
    )
    specific_humidity = mixing_ratio / (1 + mixing_ratio)
    return specific_humidity


def find_land_intersection(
    mask: xr.DataArray, land_mask: Optional[xr.DataArray] = None
) -> xr.DataArray:
    """Find points where a data mask intersects with a land mask.

    Args:
        mask: a boolean mask of data locations that includes latitude and longitude
        land_mask: a boolean mask of land locations

    Returns:
        a mask of points where AR overlaps with land
    """
    if land_mask is None:
        land_mask = regionmask.defined_regions.natural_earth_v5_0_0.land_110.mask(
            mask.longitude, mask.latitude
        )
        land_mask = land_mask.where(np.isnan(land_mask), 1).where(land_mask == 0, 0)

    # Use the scores.categorical library to compute the binary mask (true positives)
    contingency_manager = categorical.BinaryContingencyManager(mask, land_mask)
    # return the true positive mask, where mask is true and land is true
    return contingency_manager.tp


def dewpoint_from_specific_humidity(pressure: float, specific_humidity: float) -> float:
    r"""Calculate dewpoint from specific humidity.

    This computation follows the methodology used in
    `metpy.calc.dewpoint_from_specific_humidity`. Given specific humidity $q$, mixing
    ratio $w$, and pressure $p$, we compute first $w = q / (1 - q)$ and
    $e = p w / (w + \epsilon)$ where $\epsilon=0.622$ is the ratio of the molecular
    weights of water and dry air and $e$ is the partial pressure of water vapor.

    Then, we invert the Bolton (1980) formula to get the dewpoint $T_d$ given $e$:

    $$T_d =\frac{243.5 \log(e / 6.112)}{17.67 - \log(e / 6.112)}$$

    Args:
        pressure: The ambient pressure in hPa.
        specific_humidity: The specific humidity in kg/kg.

    Returns:
        The dewpoint in Kelvin.
    """
    # NOTE: there are some pathological cases where the specific humidity is negative
    # in some NWP and MLWP data sources. For safety, you should restrict the specific
    # humidity to the interval [1e-10, 1], which addresses this and should preserve the
    # remaining physical constraints.
    w = specific_humidity / (1.0 - specific_humidity)
    e = pressure * w / (w + epsilon)
    T_d = 243.5 * np.log(e / sat_press_0c) / (17.67 - np.log(e / sat_press_0c))
    return T_d + 273.15


def find_landfalls(
    track_data: xr.DataArray,
    land_geom: Optional[shapely.geometry.Polygon] = None,
    return_next_landfall: bool = False,
) -> xr.DataArray:
    """Find landfall point(s) where tracked object intersects land.

    Generalized landfall detection for any object (TC, AR, etc).
    Expects DataArray with latitude, longitude, valid_time as coords.

    Args:
        track_data: Track DataArray with latitude, longitude,
            valid_time coords. Data values are interpolated at landfall.
            Shape: (valid_time,) or (lead_time, valid_time)
        return_next_landfall: If True, return next landfall; else only first landfall
        land_geom: Shapely geometry for land intersection testing

    Returns:
        DataArray with landfall values and lat/lon/time as coords,
        or None if no landfall found
    """
    # If no land geometry is provided, use default
    if land_geom is None:
        land_geom = utils.load_land_geometry()

    # Squeeze track dimension if needed
    if "track" in track_data.dims and track_data.sizes["track"] == 1:
        track_data = track_data.squeeze("track", drop=True)

    # Detect forecast vs single track data
    is_forecast = "lead_time" in track_data.dims and "valid_time" in track_data.dims

    if is_forecast:
        # Convert to init_time coords for boundary detection
        track_data = utils.convert_valid_time_to_init_time(track_data)

        # Flatten to single time dimension for vectorized processing
        track_data_flat = track_data.stack(time=("init_time", "lead_time"))

        # Vectorized landfall detection
        landfall_mask = _detect_landfalls_wrapper(track_data_flat, land_geom)

        # Mask init_time boundaries
        landfall_mask = _mask_init_time_boundaries(landfall_mask, track_data_flat)

        # Interpolate at landfall points
        result = _interpolate_and_format_landfalls(
            track_data_flat,
            landfall_mask,
            land_geom,
            return_next_landfall,
            group_by="init_time",
        )

        return result
    else:
        # Process single track data
        landfall_mask = _detect_landfalls_wrapper(track_data, land_geom)

        return _interpolate_and_format_landfalls(
            track_data,
            landfall_mask,
            land_geom,
            return_next_landfall,
            group_by=None,
        )


def find_next_landfall_for_init_time(
    forecast_landfalls: xr.DataArray,
    target_landfalls: xr.DataArray,
) -> Optional[xr.DataArray]:
    """Find unique next upcoming landfalls from target data.

    For each forecast initialization time, finds the next landfall
    event in the target data that occurs after that init_time. Only
    unique landfalls are returned (multiple init_times mapping to the
    same landfall are deduplicated). Init_times without future
    landfalls are excluded.

    Args:
        forecast_landfalls: Forecast landfall DataArray with init_time
            dimension (from find_landfalls)
        target_landfalls: Target landfall DataArray from find_landfalls
            with return_all=True (has landfall dimension)

    Returns:
        DataArray with unique next landfalls indexed by init_time, or
        None if no future landfalls exist
    """
    # Extract init_times from forecast landfalls
    if "init_time" in forecast_landfalls.dims:
        init_times = forecast_landfalls.init_time.values
    elif "init_time" in forecast_landfalls.coords:
        # init_time as coordinate (scalar case)
        init_times = np.array([forecast_landfalls.coords["init_time"].values])
    else:
        raise KeyError("Missing init_time dimension or coordinate.")

    # Get target landfall times
    target_times = target_landfalls.coords["valid_time"].values

    # Use searchsorted to find next landfall index for each init_time
    next_indices = np.searchsorted(target_times, init_times, side="right")

    # Create mask for init_times with future landfalls
    valid_mask = next_indices < len(target_times)

    # Build result for all init_times, keeping only unique landfalls
    results = []

    for i, init_t in enumerate(init_times):
        if valid_mask[i]:
            # There is a future landfall for this init_time
            landfall_idx = next_indices[i]
            landfall = target_landfalls.isel(landfall=landfall_idx)
            landfall = landfall.assign_coords(init_time=init_t)

            results.append(landfall)

    # Combine landfalls indexed by init_time or return None if no results
    return (
        xr.concat(
            results, dim="init_time", coords="different", compat="equals", join="outer"
        )
        if results
        else None
    )


def _detect_landfalls_wrapper(
    track_data: xr.DataArray,
    land_geom: shapely.geometry.Polygon,
) -> xr.DataArray:
    """Landfall detection across consecutive point pairs.

    Args:
        track_data: Track data with latitude, longitude coords
        land_geom: Land geometry for intersection testing

    Returns:
        Boolean DataArray where True indicates a landfall between
        point i and point i+1
    """
    # Determine time dimension name
    time_dims = [d for d in track_data.dims if "time" in d.lower()]
    if not time_dims:
        return xr.DataArray(
            np.array([], dtype=bool),
            dims=track_data.dims,
        )
    time_dim = time_dims[0]

    # Extract coordinates
    lats = track_data.coords["latitude"]
    lons = track_data.coords["longitude"]

    # Convert to -180/180
    lons_180 = (lons + 180) % 360 - 180

    # Get shifted versions for consecutive pairs
    lats_next = lats.shift({time_dim: -1})
    lons_180_next = lons_180.shift({time_dim: -1})

    # Vectorize the landfall check function
    def _check_landfall_scalar(lon1, lat1, lon2, lat2):
        """Scalar landfall check for vectorization."""
        # Handle NaN values (from shift at boundaries)
        if np.isnan(lon1) or np.isnan(lat1) or np.isnan(lon2) or np.isnan(lat2):
            return False
        return _is_true_landfall(lon1, lat1, lon2, lat2, land_geom)

    # Apply to get landfall mask
    landfall_mask = xr.apply_ufunc(
        _check_landfall_scalar,
        lons_180,
        lats,
        lons_180_next,
        lats_next,
        vectorize=True,  # We already vectorized with np.vectorize
        dask="parallelized",
        output_dtypes=[bool],
    )

    return landfall_mask


def _mask_init_time_boundaries(
    landfall_mask: xr.DataArray,
    track_data: xr.DataArray,
) -> xr.DataArray:
    """Mask out landfalls at init_time boundaries.

    Prevents comparing last point of one init_time with
    first point of next init_time.

    Args:
        landfall_mask: Boolean mask of potential landfalls
        track_data: Original track data with init_time coord

    Returns:
        Masked landfall array with boundaries set to False
    """
    if "init_time" not in track_data.coords:
        return landfall_mask

    # Determine time dimension
    time_dims = [d for d in track_data.dims if "time" in d.lower()]
    if not time_dims:
        return landfall_mask
    time_dim = time_dims[0]

    # Get init_time for each point and the next point
    init_curr = track_data.coords["init_time"]
    init_next = init_curr.shift({time_dim: -1})

    # Only keep landfalls where init_time doesn't change
    same_init = init_curr == init_next

    # Also handle NaN from shift operation
    same_init = same_init.fillna(False)

    return landfall_mask & same_init


def _interpolate_and_format_landfalls(
    track_data: xr.DataArray,
    landfall_mask: xr.DataArray,
    land_geom: shapely.geometry.Polygon,
    return_all_landfalls: bool,
    group_by: Optional[str] = None,
) -> Optional[xr.DataArray]:
    """Interpolate landfall points and format output.

    Args:
        track_data: Original track data
        landfall_mask: Boolean mask of landfall locations
        land_geom: Land geometry for intersection
        return_all_landfalls: Whether to return all or just first
        group_by: If not None, group by this coord (e.g., "init_time")

    Returns:
        Formatted landfall DataArray or None if no landfalls
    """
    # Determine time dimension
    time_dims = [d for d in track_data.dims if "time" in d.lower()]
    if not time_dims:
        return None

    # Get indices where landfalls occur
    landfall_indices = np.where(landfall_mask.values)[0]

    if len(landfall_indices) == 0:
        return None

    # Extract coordinate arrays
    lats_vals = track_data.coords["latitude"].values
    lons_vals = track_data.coords["longitude"].values
    times_vals = track_data.coords["valid_time"].values
    track_vals = track_data.values

    # Convert longitudes to -180/180
    lons_180 = (lons_vals + 180) % 360 - 180

    # Get init_time if available
    init_times = None
    if "init_time" in track_data.coords:
        init_times = track_data.coords["init_time"].values

    # Process each landfall
    landfall_data = []

    for i in landfall_indices:
        try:
            # Create segment
            segment = shapely.geometry.LineString(
                [(lons_180[i], lats_vals[i]), (lons_180[i + 1], lats_vals[i + 1])]
            )

            # Get intersection point
            intersection = segment.intersection(land_geom)

            # Handle different intersection types
            if intersection.is_empty:
                continue

            # Get first intersection point based on geometry type
            geom_type = intersection.geom_type

            if geom_type == "Point":
                landfall_lon, landfall_lat = intersection.x, intersection.y
            elif geom_type == "LineString":
                landfall_lon, landfall_lat = intersection.coords[0]
            elif geom_type in ("MultiPoint", "MultiLineString", "GeometryCollection"):
                # Get first geometry from collection
                first_geom = intersection.geoms[0]
                if first_geom.geom_type == "Point":
                    landfall_lon, landfall_lat = first_geom.x, first_geom.y
                else:
                    landfall_lon, landfall_lat = first_geom.coords[0]
            else:
                continue

            # Interpolate
            full_dist = segment.length
            if full_dist == 0:
                continue

            landfall_dist = shapely.geometry.LineString(
                [(lons_180[i], lats_vals[i]), (landfall_lon, landfall_lat)]
            ).length
            frac = landfall_dist / full_dist

            landfall_point = {
                "latitude": landfall_lat,
                "longitude": utils.convert_longitude_to_360(landfall_lon),
                "valid_time": times_vals[i]
                + frac * (times_vals[i + 1] - times_vals[i]),
                "value": track_vals[i] + frac * (track_vals[i + 1] - track_vals[i])
                if not np.isnan(track_vals[i])
                else np.nan,
            }

            # Add init_time if available
            if init_times is not None:
                landfall_point["init_time"] = init_times[i]

            landfall_data.append(landfall_point)

        except (IndexError, AttributeError, ValueError, TypeError, ZeroDivisionError):
            continue

    if not landfall_data:
        return None

    # Check if all landfall values are NaN
    all_nan = all(np.isnan(d.get("value", np.nan)) for d in landfall_data)
    if all_nan:
        return None

    # Format output based on grouping
    if group_by == "init_time" and init_times is not None:
        # Group by init_time and optionally keep first per group
        results = []
        unique_inits = np.unique([d["init_time"] for d in landfall_data])

        for init_t in unique_inits:
            init_landfalls = [d for d in landfall_data if d["init_time"] == init_t]

            if not return_all_landfalls:
                # Keep only first landfall for this init_time
                init_landfalls = sorted(init_landfalls, key=lambda x: x["valid_time"])
                init_landfalls = [init_landfalls[0]]

            # Create DataArray with consistent landfall dimension
            values = [d["value"] for d in init_landfalls]
            coords = {
                "latitude": (["landfall"], [d["latitude"] for d in init_landfalls]),
                "longitude": (["landfall"], [d["longitude"] for d in init_landfalls]),
                "valid_time": (["landfall"], [d["valid_time"] for d in init_landfalls]),
                "landfall": np.arange(len(init_landfalls)),
                "init_time": init_t,
            }
            da = xr.DataArray(
                values,
                dims=["landfall"],
                coords=coords,
                name=track_data.name or "landfall_value",
            )
            results.append(da)

        return (
            xr.concat(
                results,
                dim="init_time",
                coords="different",
                compat="equals",
                join="outer",
            )
            if results
            else None
        )

    else:
        # No grouping - format as single or multiple landfalls
        if return_all_landfalls:
            # Return with landfall dimension
            values = [d["value"] for d in landfall_data]
            coords = {
                "latitude": (["landfall"], [d["latitude"] for d in landfall_data]),
                "longitude": (["landfall"], [d["longitude"] for d in landfall_data]),
                "valid_time": (["landfall"], [d["valid_time"] for d in landfall_data]),
                "landfall": np.arange(len(landfall_data)),
            }
            return xr.DataArray(
                values,
                dims=["landfall"],
                coords=coords,
                name=track_data.name or "landfall_value",
            )
        else:
            # Return first landfall as scalar
            d = landfall_data[0]
            coords = {
                "latitude": d["latitude"],
                "longitude": d["longitude"],
                "valid_time": d["valid_time"],
            }
            if "init_time" in d:
                coords["init_time"] = d["init_time"]
            return xr.DataArray(
                d["value"],
                coords=coords,
                name=track_data.name or "landfall_value",
            )


# Keep existing _process_single_track_landfall for backward compatibility
# or remove if fully replaced


def _is_true_landfall(
    lon1: float,
    lat1: float,
    lon2: float,
    lat2: float,
    land_geom: shapely.geometry.Polygon,
) -> bool:
    """Detect true landfall (ocean to land movement).

    This function is a way to detect if a track crosses land. There are a few scenarios,
    being ocean to land, ocean to ocean, and land to ocean. Ocean to ocean can have a
    landfall if there is land between the two points.

    Args:
        lon1, lat1: Starting point coordinates
        lon2, lat2: Ending point coordinates
        land_geom: Land geometry for intersection testing

    Returns:
        True if this represents a landfall, False otherwise
    """
    try:
        start_point = shapely.geometry.Point(lon1, lat1)
        end_point = shapely.geometry.Point(lon2, lat2)

        start_over_land = land_geom.contains(start_point)
        end_over_land = land_geom.contains(end_point)

        # Ocean -> Land = LANDFALL
        if not start_over_land and end_over_land:
            return True

        # Ocean -> Ocean, check if land is between
        if not start_over_land and not end_over_land:
            segment = shapely.geometry.LineString([(lon1, lat1), (lon2, lat2)])
            if segment.intersects(land_geom):
                return True

        # Land -> Ocean or Ocean -> Ocean (no intersection) = NOT LANDFALL
        return False

    except (AttributeError, ValueError, TypeError):
        # Return False if geometry operations fail:
        # - AttributeError: invalid/None geometry
        # - ValueError/TypeError: invalid coordinate values
        return False
